/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright held by original author
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

Application
    pvCoupledFoam

Description
    Steady-state solver for incompressible, turbulent flow, with implicit
    coupling between pressure and velocity achieved by BlockLduMatrix
    Turbulence is in this version solved using the existing turbulence
    structure.
Auther
    Klas Jareteg, 2012-09-27-2012-10-16
Version
    0.6 

\*---------------------------------------------------------------------------*/

#include "fvCFD.H"
#include "singlePhaseTransportModel.H"
#include "RASModel.H"


// Add the support for arbitry vector size types
#include "VectorNFieldTypes.H"
#include "volVectorNFields.H"

// Add the support for block matrices
#include "blockLduSolvers.H"
#include "blockVectorNMatrices.H"

// Add the utilities written for blockMatrix transformations
#include "blockMatrixTools.H"

// Surface interpolation schemes to get the weights
#include "surfaceInterpolationScheme.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

int main(int argc, char *argv[])
{
#   include "setRootCase.H"
#   include "createTime.H"

Info<< "Create mesh for time = "
    << runTime.timeName() << Foam::nl << Foam::endl;
Info<<" Mesh name: "<<fvMesh::defaultRegion<<endl;
Foam::fvMesh mesh
(
    Foam::IOobject
    (
        Foam::fvMesh::defaultRegion,
        //"fluid",
        runTime.timeName(),
        runTime,
        Foam::IOobject::MUST_READ
    )
);

Info << "Reading field p\n" << endl;
volScalarField p
(
    IOobject
    (
        "p",
        runTime.timeName(),
        mesh,
        IOobject::MUST_READ,
        IOobject::AUTO_WRITE
    ),
    mesh
);

Info << "Reading field U\n" << endl;
volVectorField U
(
    IOobject
    (
        "U",
        runTime.timeName(),
        mesh,
        IOobject::MUST_READ,
        IOobject::AUTO_WRITE
    ),
    mesh
);

#include "createPhi.H"

singlePhaseTransportModel laminarTransport(U, phi);
autoPtr<incompressible::RASModel> turbulence
(
    incompressible::RASModel::New(U, phi, laminarTransport)
);

// Block vector field for the pressure and velocity field to be solved for
volVector4Field pU
(
    IOobject
    (
       "pU",
       runTime.timeName(),
       mesh,
       IOobject::NO_READ,
       IOobject::NO_WRITE
    ),
    mesh,
    dimensionedVector4(word(),dimless,vector4::zero)
);

// Insert the pressure and velocity internal fields in to the volVector2Field
{
    vector4Field blockX = pU.internalField();

    // Separately add the three velocity components
    for (int i=0; i<3;i++)
    {
        tmp<scalarField> tf = U.internalField().component(i);
        scalarField& f = tf();
        blockMatrixTools::blockInsert(i,f,blockX);
    }

    // Pressure is the 2nd component
    scalarField& f = p.internalField();
    blockMatrixTools::blockInsert(3,f,blockX);
}

// ===========================================================================//

    Info<< "\nStarting time loop\n" << endl;
    while (runTime.loop())
    {
        Info<< "Time = " << runTime.timeName() << nl << endl;

        p.storePrevIter();
        U.storePrevIter();

        const surfaceVectorField& Sf = p.mesh().Sf();
        const unallocLabelList& owner = mesh.owner();
        const unallocLabelList& neighbour = mesh.neighbour();

        {
            vector4Field blockX = pU.internalField();
        
            // Separately add the three velocity components
            for (int i=0; i<3;i++)
            {
                tmp<scalarField> tf = U.internalField().component(i);
                scalarField& f = tf();
                blockMatrixTools::blockInsert(i,f,blockX);
            }
        
            // Pressure is the 2nd component
            scalarField& f = p.internalField();
            blockMatrixTools::blockInsert(3,f,blockX);


// ===========================================================================//
// Velocity equation (LHS) matrix
// ===========================================================================//
            tmp<fvVectorMatrix> UEqnLHS
            (
                fvm::div(phi,U)
              + turbulence->divDevReff(U)
            );

            // Matrix block 
            BlockLduMatrix<vector4> B(mesh);
            
            // Diagonal is set separately 
            Field<tensor4>& d = B.diag().asSquare();

            // Off-diagonal also as square
            Field<tensor4>& u = B.upper().asSquare();
            Field<tensor4>& l = B.lower().asSquare();

            // Source term for the block matrix
            Field<vector4> s(mesh.nCells(), vector4::zero);
            
            // Add the boundary contributions for the velocity equation
            tmp<scalarField> tdiag = UEqnLHS().D();
            scalarField& diag = tdiag();
            scalarField& upper = UEqnLHS().upper();
            scalarField& lower = UEqnLHS().lower();
            
            // Add diagonal boundary contribution
            // This is automatically done when you do UEqnLHS().D();
            //UEqnLHS().addBoundaryDiag(diag,0);

            // Add source boundary contribution
            vectorField& source = UEqnLHS().source();             
            UEqnLHS().addBoundarySource(source, false);
           
// ===========================================================================//
// Pressure gradient matrix
// ===========================================================================//
            // Interpolation scheme for the pressure weights
            tmp<surfaceInterpolationScheme<scalar> > 
            tinterpScheme_
            (
                surfaceInterpolationScheme<scalar>::New
                (
                    p.mesh(),
                    p.mesh().interpolationScheme("grad(p)")
                )
            );

            // Pressure gradient contributions - corresponds to an implicit 
            // gradient operator
            tmp<vectorField> tpUv = tmp<vectorField>
                (
                    new vectorField(upper.size(),pTraits<vector>::zero)
                );
            vectorField& pUv = tpUv();
            tmp<vectorField> tpLv = tmp<vectorField>
                (
                    new vectorField(lower.size(),pTraits<vector>::zero)
                );
            vectorField& pLv = tpLv();
            tmp<vectorField> tpSv = tmp<vectorField>
                (
                    new vectorField(source.size(),pTraits<vector>::zero)
                );
            vectorField& pSv = tpSv();
            tmp<vectorField> tpDv = tmp<vectorField>
                (
                    new vectorField(diag.size(),pTraits<vector>::zero)
                );
            vectorField& pDv = tpDv();
            
            // 2) Use interpolation weights to assemble the contributions
            tmp<surfaceScalarField> tweights = tinterpScheme_().weights(p);
            const surfaceScalarField& weights = tweights();
            
            for(int i=0;i<owner.size();i++)
            {
                int o = owner[i];
                int n = neighbour[i];
                scalar w = weights.internalField()[i];
                vector s = Sf[i];

                pDv[o]+=s*w;
                pDv[n]-=s*(1-w);
                pLv[i]=-s*w;
                pUv[i]=s*(1-w);
            
            }

            // Get boundary condition contributions for pressure grad(P) 
            p.boundaryField().updateCoeffs();
            forAll(p.boundaryField(),patchI)
            {
                // Present fvPatchField
                fvPatchField<scalar> & fv = p.boundaryField()[patchI];

                // Retrieve the weights for the boundary
                const fvsPatchScalarField& pw = weights.boundaryField()[patchI];

                // Contributions from the boundary coefficients 
                tmp<Field<scalar> > tic = fv.valueInternalCoeffs(pw);
                Field<scalar>& ic = tic();
                tmp<Field<scalar> > tbc = fv.valueBoundaryCoeffs(pw);
                Field<scalar>& bc = tbc();
                
                // Get the fvPatch only
                const fvPatch& patch = fv.patch();
                
                // Surface normals for this patch
                tmp<Field<vector> > tsn = patch.Sf();
                Field<vector> sn = tsn();

                // Manually add the contributions from the boundary
                // This what happens with addBoundaryDiag, addBoundarySource
                forAll(fv,facei)
                {
                    label c = patch.faceCells()[facei];
                    
                    pDv[c]+=ic[facei]*sn[facei];
                    pSv[c]-=bc[facei]*sn[facei];
                }
            }

// ===========================================================================//
// Assemble momentum equation
// ===========================================================================//
            // Assemble the momentum equation contributions
            forAll(d,i)
            {
                d[i](0,0) = diag[i]; 
                d[i](1,1) = diag[i];
                d[i](2,2) = diag[i];
                
                d[i](0,3) = pDv[i].x();
                d[i](1,3) = pDv[i].y();
                d[i](2,3) = pDv[i].z();
            }
            forAll(l,i)
            {
                l[i](0,0) = lower[i];
                l[i](1,1) = lower[i];
                l[i](2,2) = lower[i];
                
                l[i](0,3) = pLv[i].x();
                l[i](1,3) = pLv[i].y();
                l[i](2,3) = pLv[i].z();
            }
            forAll(u,i)
            {
                u[i](0,0) = upper[i];
                u[i](1,1) = upper[i];
                u[i](2,2) = upper[i];

                u[i](0,3) = pUv[i].x();
                u[i](1,3) = pUv[i].y();
                u[i](2,3) = pUv[i].z();
            }
            forAll(s,i)
            {
                s[i](0) = source[i].x()+pSv[i].x();
                s[i](1) = source[i].y()+pSv[i].y();
                s[i](2) = source[i].z()+pSv[i].z();
            }

// ===========================================================================//
// Create implicit velocity (LHS) for continuity equation
// ===========================================================================//
            // Again an implicit version not existing, now the div operator
            tmp<surfaceInterpolationScheme<scalar> > 
            UtinterpScheme_
            (
                surfaceInterpolationScheme<scalar>::New
                (
                    U.mesh(),
                    U.mesh().interpolationScheme("div(U)(implicit)")
                )
            );
            

            // 1) Setup diagonal, source, upper and lower
            tmp<vectorField> tMUpper = tmp<vectorField> 
                (new vectorField(upper.size(),pTraits<vector>::zero));
            vectorField& MUpper = tMUpper();
            
            tmp<vectorField> tMLower = tmp<vectorField> 
                (new vectorField(lower.size(),pTraits<vector>::zero));
            vectorField& MLower = tMLower();
            
            tmp<vectorField> tMDiag = tmp<vectorField> 
                (new vectorField(diag.size(),pTraits<vector>::zero));
            vectorField& MDiag = tMDiag();

            tmp<vectorField> tMSource = tmp<vectorField> 
                (
                    new vectorField
                    (
                        source.component(0)().size(),pTraits<vector>::zero
                    )
                );
            vectorField& MSource = tMSource();
            
            // 2) Use interpolation weights to assemble the contributions
            tmp<surfaceScalarField> tMweights = 
                UtinterpScheme_().weights(mag(U));
            const surfaceScalarField& Mweights = tMweights();

            for(int i=0;i<owner.size();i++)
            {
                int o = owner[i];
                int n = neighbour[i];
                scalar w = Mweights.internalField()[i];
                vector s = Sf[i];
    
                MDiag[o]+=s*w;
                MDiag[n]-=s*(1-w);
                MLower[i]=-s*w;
                MUpper[i]=s*(1-w);
            }

            // Get boundary condition contributions for the pressure grad(P) 
            U.boundaryField().updateCoeffs();
            forAll(U.boundaryField(),patchI)
            {
                // Present fvPatchField
                fvPatchField<vector> & fv = U.boundaryField()[patchI];

                // Retrieve the weights for the boundary
                const fvsPatchScalarField& Mw = 
                    Mweights.boundaryField()[patchI];

                // Contributions from the boundary coefficients 
                tmp<Field<vector> > tic = fv.valueInternalCoeffs(Mw);
                Field<vector>& ic = tic();
                tmp<Field<vector> > tbc = fv.valueBoundaryCoeffs(Mw);
                Field<vector>& bc = tbc();
                
                // Get the fvPatch only
                const fvPatch& patch = fv.patch();
                
                // Surface normals for this patch
                tmp<Field<vector> > tsn = patch.Sf();
                Field<vector> sn = tsn();

                // Manually add the contributions from the boundary
                // This what happens with addBoundaryDiag, addBoundarySource
                forAll(fv,facei)
                {
                    label c = patch.faceCells()[facei];
                   
                    MDiag[c]+=cmptMultiply(ic[facei],sn[facei]);
                    MSource[c]-=cmptMultiply(bc[facei],sn[facei]);
                }
            }

// ===========================================================================//
// Create explicit and implict pressure parts for continuity equation
// ===========================================================================//
            // Pressure parts of the mass equation
            tmp<volScalarField> tA = UEqnLHS().A();
            volScalarField& A = tA();

            tmp<volVectorField> texp = fvc::grad(p);
            volVectorField& exp = texp();
            tmp<volVectorField> texp2 = exp/A;
            volVectorField exp2 = texp2();
            
            tmp<fvScalarMatrix> MEqnLHSp
            (
                -fvm::laplacian(1/A,p)
                ==
                -fvc::div(exp2)
            );
            
            // Add the boundary contributions
            scalarField& pMdiag = MEqnLHSp().diag();
            scalarField& pMupper = MEqnLHSp().upper();
            scalarField& pMlower = MEqnLHSp().lower();
            
            // Add diagonal boundary contribution
            MEqnLHSp().addBoundaryDiag(pMdiag,0);

            // Add source boundary contribution
            scalarField& pMsource = MEqnLHSp().source();             
            MEqnLHSp().addBoundarySource(pMsource, false);
            
// ===========================================================================//
// Assemble mass equation
// ===========================================================================//
            forAll(d,i)
            {
                d[i](3,0) = MDiag[i].x(); 
                d[i](3,1) = MDiag[i].y(); 
                d[i](3,2) = MDiag[i].z(); 
                d[i](3,3) = pMdiag[i];
            }
            forAll(l,i)
            {
                l[i](3,0) = MLower[i].x();
                l[i](3,1) = MLower[i].y();
                l[i](3,2) = MLower[i].z();
                l[i](3,3) = pMlower[i];
            }
            forAll(u,i)
            {
                u[i](3,0) = MUpper[i].x();
                u[i](3,1) = MUpper[i].y();
                u[i](3,2) = MUpper[i].z();
                u[i](3,3) = pMupper[i];
            }
            forAll(s,i)
            {
                s[i](3) =  MSource[i].x()
                          +MSource[i].y()
                          +MSource[i].z()
                          +pMsource[i];
            }
            
// ===========================================================================//
// Solve the block matrix
// ===========================================================================// 
            BlockSolverPerformance<vector4> solverPerf = 
                BlockLduSolver<vector4>::New
                (
                    word("blockVar"),
                    B,
                    mesh.solver("blockVar")
                )->solve(pU,s);

            solverPerf.print();

// ===========================================================================//
// Retrieve the solution and update for next iteration
// ===========================================================================//
             
            tmp<scalarField> tUx = U.internalField().component(0);
            scalarField& Ux = tUx();
            blockMatrixTools::blockRetrieve(0, Ux, pU);
            U.internalField().replace(0,Ux);
            
            tmp<scalarField> tUy = U.internalField().component(1);
            scalarField& Uy = tUy();
            blockMatrixTools::blockRetrieve(1, Uy, pU);
            U.internalField().replace(1,Uy);
            
            tmp<scalarField> tUz = U.internalField().component(2);
            scalarField& Uz = tUz();
            blockMatrixTools::blockRetrieve(2, Uz, pU);
            U.internalField().replace(2,Uz);
            
            blockMatrixTools::blockRetrieve(3, p.internalField(), pU);
            
            UEqnLHS.clear();
            
            p.relax();
            
            U.correctBoundaryConditions();
            p.correctBoundaryConditions();

        }

        phi = fvc::interpolate(U) & mesh.Sf();
        turbulence->correct();
        runTime.write();

        Info<< "ExecutionTime = " << runTime.elapsedCpuTime() << " s"
            << "  ClockTime = " << runTime.elapsedClockTime() << " s"
            << nl << endl;

    }

    Info<< "End\n" << endl;

    return 0;
}
